/*  $Id: Elf32_dump.cc,v 1.2 2000/01/12 03:44:44 shaggy Exp $
 *      relative path: lib/bff/elf32
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the ShagOS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/members/shaggy
 */

#pragma Make Includes_Banned += -nostdinc -nostdinc++ 
#pragma Make Flags_Banned += -static 

#define USE_STDLIB
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <vstream>
#include <vdir>
#include <vthread>
#include <Error>
#include "Elf32.hh"


const char*
Elf32::shstr(Word index) const
{
	if(!shstrtab && ehdr.e_shstrndx) {
		Shdr shdr;
		read(shdr, ehdr.e_shstrndx);
		const_cast<char*&>(shstrtab)
			= new char[shdr.sh_size];
		stream->read_at(shdr.sh_offset, shstrtab, shdr.sh_size);
		return &shstrtab[index];
	} 
	if(shstrtab)
		return &shstrtab[index];
	else
		return "?";
}


const char*
Elf32::section(int index) const
{
	Shdr shdr;

	switch(index) {
	case SHN_UNDEF:
		return "*UND*"; 
	case SHN_ABS:
		return "*ABS*";
	case SHN_COMMON:
		return "*COM*";
	default:
		if(index >= SHN_LORESERVE)
			return "*RSV*";
		read(shdr, index);
		return shstr(shdr.sh_name);
	}
}



void
Elf32::dump(const Ehdr &e)
{
	printf("ELF %s %s v%d  p:%lx,%x  s:%lx,%x f:%lx\n",
		ei_class_str(e.e_ident[EI_CLASS]),
		ei_data_str(e.e_ident[EI_DATA]),
		e.e_ident[EI_VERSION],
		e.e_phoff, e.e_phentsize,
		e.e_shoff, e.e_shentsize, e.e_flags);
	printf("ET:%s EM:%s v:%ld entry:%08lx p:%d s:%d/%d\n",
	        e_type_str(e.e_type), e_machine_str(e.e_machine),
		e.e_version, e.e_entry, e.e_phnum, e.e_shnum,
		e.e_shstrndx);
}

void
Elf32::dump(const Phdr &p)
{
	printf("PT:%s v:%08lx p:%08lx s:%lx o:%lx s:%lx f:%lx a:%ld\n",
		p_type_str(p.p_type), p.p_vaddr, p.p_paddr,
		p.p_memsz, p.p_offset, p.p_filesz,
		p.p_flags, p.p_align);
}

void
Elf32::dump(const Shdr &s) const
{
	printf("[%s] SHT:%s %c%c%c a:%08lx o:%lx s:%lx l:%ld a:%ld e:%ld\n",
		shstr(s.sh_name), sh_type_str(s.sh_type),
		(s.sh_flags&SHF_ALLOC)?'a':'-',
		(s.sh_flags&SHF_WRITE)?'w':'-',
		(s.sh_flags&SHF_EXECINSTR)?'x':'-',
		s.sh_addr, s.sh_offset, s.sh_size,
		s.sh_link, s.sh_addralign, s.sh_entsize);
}


void
Elf32::dump(const Sym &s, const char *strtab) const
{
	printf("%08lx/%lx %c %c %s %s\n",
		s.st_value, s.st_size,
		"LGW             "[s.st_type()],
		"?OCSF           "[s.st_bind()],
		section(s.st_shndx),
		s.st_name?&strtab[s.st_name]:"?");
}


void
Elf32::dump(const Rel &r, const Sym *sym, const char *strtab) const
{
	printf("%08lx %s : ",
		r.r_offset, r_type_str(r.r_type()));
	dump(sym[r.r_sym()], strtab);
}

void
Elf32::dump(const Rela &r, const Sym *sym, const char *strtab) const
{
	printf("%08lx %s %lx: ",
		r.r_offset, r_type_str(r.r_type()),
		r.r_addend);
	dump(sym[r.r_sym()], strtab);
}


void
Elf32::dump() const
{
	dump(ehdr);
	for(int i=0; i<ehdr.e_phnum; i++) {
		Phdr phdr;
		read(phdr, i);
		printf("ph %d: ", i);
		dump(phdr);
	}
	for(int i=0; i<ehdr.e_shnum; i++) {
		Shdr shdr;
		read(shdr, i);
		printf("sh %d: ", i);
		dump(shdr);
	}
}


void
Elf32::dyn_chain::dump(int indent)
{
	for(int i=0; i<indent*4; i++)
		printf(" ");
	printf("%s (link=%p,load=%p,dyn=%p) %d\n", link->libname,
			link, (void*)(link->loadaddr), link->dynamic_addr,
			link->initialized);
	if(next_level)
		next_level->dump(indent+1);
	if(next)
		next->dump(indent);
}


void
Elf32::link_info::dump()
{
	printf("link_info for %s (%p)\n", libname, this);
	for(const Dyn *dp = dynamic_addr; dp->d_tag != DT_NULL; dp++) {
		const Dyn &d = *dp;
		printf("%08lx %s: ",
			d.d_val, d_tag_str(d.d_tag));
		switch(d.d_tag) {
		case DT_NEEDED:
		case DT_SONAME:
		case DT_RPATH:
			printf("%s", &strtab[d.d_val]);
		}
		printf("\n");
	}
}

