/*  $Id: mksolo.cc,v 1.7 2000/01/17 23:01:39 shaggy Exp $
 *      configure a disk to use SOLO
 *      
 *      relative path: sys/boot/m-pc
 *  
 *  Copyright (c) 1994-1999 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the ShagOS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/members/shaggy
 */

#define mksolo_VER_MAJOR	0
#define mksolo_VER_MINOR	30
#define mksolo_VER_PATCH	0
#define mksolo_VER_STR	"0.30"
#define mksolo_VER_DATE	"January 12, 2000"
#define mksolo_VER_NDATE	"01/12/00"
 
#define _POSIX_SOURCE
#include <libc.h>
#include <stdx1.h>
#include <dirent.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>

#include <sys/ioctl.h>
#include <sys/stat.h>
#include <linux/major.h>
#include <linux/fs.h>
#include <linux/fd.h>
#include <linux/hdreg.h>
#include <boot/mapfile.h>

#include <iostream.h>
#include <fstream.h>
#include <new.h>

#include "smstring.hh"
#include "ArgParse.h"

#include "../boot.h"
#include "../device.h"

#include "diskmap.h"


void
find_device(string* dname, csz* fname)
{
	sz d[_POSIX_PATH_MAX];
	struct stat stbuf;

	if(stat(fname, &stbuf) < 0) {
		perror(fname);
		exit(1);
	}
	if(devname(stbuf.st_dev, d) < 0) {
		cerr << "Cannot find device " << hex << stbuf.st_dev
			<< " in /dev" << endl; 
		exit(1);
	}
	*dname = d;
}

ostream&
operator<<(ostream &os, string s)
{
	os << s.chars();
	return os;
}


bool debug = false;


static csz rcsid[] __attribute__((unused))
	= "$Id: mksolo.cc,v 1.7 2000/01/17 23:01:39 shaggy Exp $"
	  "$Release: " mksolo_VER_STR " $";


class File_Geometry 
{
	struct diskinfo di;
	struct diskmap dm;
	int	fd;
	bool	compact;
	uint32	mtime_hi;
	bool	linear;
	int	partition;
	string	name;
public:
	File_Geometry(const string fname, bool set_compact = false);
	~File_Geometry();
	bool	next_chunk(Chunk* c);
	int	get_size()		{ return dm.size; }
	int	get_device()		{ return di.biosid; }
	int	get_unit() { return linear ? di.biosid
				|((partition|DISK_LINEAR)<<8)
				: di.biosid|(partition<<8); }
	time_t	get_mtime_hi()		{ return mtime_hi; }
};



File_Geometry::File_Geometry(string fname, bool set_compact)
{
	struct stat stbuf;

	name = fname;
	compact = set_compact;
	if((fd = open(fname.chars(), O_RDWR)) < 0) {
		perror(fname.chars());
		exit(1);
	}
	if(fstat(fd, &stbuf) < 0) {
		perror(fname.chars());
		exit(1);
	}
	// convert Unix time in seconds to ShagOS time
	// in nanoseconds
	mtime_hi = ((int64)stbuf.st_mtime*(int64)1000000000) >> 32;

	dm.pos = -1;
	if(diskmap(fd, &di, &dm) < 0) {
		perror(fname.chars());
		exit(1);
	}

	linear = false;	// use cylinder/head/sector
	partition = 0;	// whole disk
	if(!compact && di.maxmult > dm.blocksize/di.secsize)
		di.maxmult = dm.blocksize/di.secsize;
	if(debug) {
		cout << "Disk params: " << hex << di.devid << '/'
			<< di.biosid << " cylinders:" << dec << di.cyls
			<< " heads:" << di.heads << " sectors:" << di.secs
			<< " start:" << di.start 
			<< " maxmult:" << di.maxmult << endl;
		cout << "\tblocksize:" << dm.blocksize 
			<< " sectorsize:" << di.secsize
			<< " size:" << dm.size 
			<< " file: " << fname << endl;
	}
}



File_Geometry::~File_Geometry()
{
	close(fd);
}


bool
File_Geometry::next_chunk(Chunk *c)
{
	unsigned int cyl, head;
	if(dm.pos >= (off_t)dm.size)
		return false;
	if(dm.pos == 0) {
		if(diskmap(fd, &di, &dm) < 0) {
			perror("diskmap");
			exit(1);
		}
		if(dm.ofs != 0)  {
			fprintf(stderr, "dm.ofs: %d != 0\n", dm.ofs);
			exit(1);
		}
	}
	if(dm.cyl >= 1024) {
		fprintf(stderr, "\a%s: pos=%ld, cyl=%d: BIOS cannot handle cylinder >= 1024\n", name.chars(), dm.pos, dm.cyl);
		exit(1);
	}
	cyl = dm.cyl;
	c->cyl1 = cyl&0xff;
	c->cyl2 = (cyl>>8)&3;
	c->head = head = dm.head;
	c->sector = dm.sec;
	c->partial = false;
	c->count = dm.nsec;
	dm.pos += dm.len;
	while(dm.pos < (off_t)dm.size) {
		if(diskmap(fd, &di, &dm) < 0) {
			perror("diskmap");
			exit(1);
		}
		if(dm.ofs != 0)  {
			fprintf(stderr, "dm.ofs: %d != 0\n", dm.ofs);
			exit(1);
		}
		if((dm.sec == 0 && c->sector == 0 && c->count+dm.nsec < 64)
			|| (c->count+dm.nsec <= di.maxmult
				&& dm.sec == c->sector+c->count
				&& dm.cyl == cyl
				&& dm.head == head)) {
			c->count += dm.nsec;
			dm.pos += dm.len;
		} else
			break;
	}
	return true;
}




void
save_boot_block(string dev, string save_fname)
{
	struct BootBlock bootblock;
	int fd;

	ifstream savefile_exists(save_fname.chars());
	if(savefile_exists.is_open()) {
		savefile_exists.close();
		return;
	}

	if((fd = open(dev.chars(), O_RDONLY)) < 0) {
		perror(dev.chars());
		exit(1);
	}
	read(fd, &bootblock, sizeof(bootblock));

	File_Geometry geo(dev.chars());
	Chunk chunk;
	if(!geo.next_chunk(&chunk)) {
		perror("next_chunk");
		exit(1);
	}
	if(bootblock.magic == BOOT_MAGIC) {
		bootblock.magic = BOOTMERGE_MAGIC;
		*(Chunk*)bootblock.sysname = chunk;
	}
	ofstream savefile(save_fname.chars(), ios::out | ios::noreplace);
	if(!savefile.is_open()) 
		return;
	cout << "saving bootsector from '" << dev << "' at "
		<< (chunk.cyl1 | (chunk.cyl2<<8)) << "," 
		<< (int)chunk.head << "," << (int)chunk.sector << "  to '"
		<< save_fname << "'" << endl;
	savefile.write(&bootblock, sizeof(bootblock));
	close(fd);
	savefile.close();
}


void
restore_boot_block(string dev, string save_fname)
{
	struct BootBlock bootblock;
	struct BootBlock oldblock;
	int fd;

	ifstream savefile(save_fname.chars());
	if(!savefile.is_open()) {
		cout << "could not open saved bootsector '" << save_fname << "'" << endl;
		exit(1);
	}
	if((fd = open(dev.chars(), O_RDWR)) < 0) {
		perror(dev.chars());
		exit(1);
	}
	read(fd, &bootblock, sizeof bootblock);
	savefile.read(&oldblock, sizeof oldblock);
	savefile.close();
	if(oldblock.magic == BOOTMERGE_MAGIC) {
		memcpy(bootblock.jmp, oldblock.jmp, sizeof bootblock.jmp);
		memcpy(bootblock.code, oldblock.code, 
			sizeof(bootblock.code)+sizeof(bootblock.bootoffset)
			+sizeof(bootblock.chunk)
			+sizeof(bootblock.mapsize)+sizeof(bootblock.mapdev)
			+sizeof(bootblock.diskparam));
		bootblock.magic = BOOT_MAGIC;
	} else
		memcpy(&bootblock, &oldblock, sizeof bootblock);

	lseek(fd, 0, SEEK_SET);
	write(fd, &bootblock, sizeof(bootblock));
	cout << "bootsector restored from '" << save_fname << "'" << endl;
	close(fd);
}


void
write_boot_block(string dev, string bootblock_fname, File_Geometry& mapgeo, int bootoffset)
{
	struct BootBlock bootblock;
	int i;
	int fd;

	if((fd = open(dev.chars(), O_RDWR)) < 0) {
		perror(dev.chars());
		exit(1);
	}
	read(fd, &bootblock, sizeof(bootblock));

	ifstream f(bootblock_fname.chars());
	if(!f.is_open()) {
		perror(bootblock_fname.chars());
		exit(1);
	}

	bootblock.jmp[0] = 0xe9;	// JMP
	bootblock.jmp[1] = (bootblock.code-bootblock.jmp)-3;
	bootblock.jmp[2] = 0x00;
	f.read(&bootblock.code, sizeof(bootblock.code));
	f.close();
	i = 0;

	while(mapgeo.next_chunk(&bootblock.chunk[i]))
		i++;
	bootblock.mapsize = i;
	bootblock.magic = BOOT_MAGIC;
	bootblock.mapdev = mapgeo.get_device();
	bootblock.bootoffset = bootoffset;

	lseek(fd, 0, SEEK_SET);
	write(fd, &bootblock, sizeof(bootblock));
	close(fd);
	if(debug) {
		cout << "map size: " << (int)bootblock.mapsize << endl;
		for(i=0; i< bootblock.mapsize; i++) {
			Chunk &c = bootblock.chunk[i];
			cout << "head:" << c.head
			     << " cyl:"
			     << (c.cyl1 | (c.cyl2<<8)) 
			     << " sec:" << c.sector;
			if(c.partial) {
				if(c.count&0x40)	/* small */
					cout << " offset:" 
					     << c.count*16
					     << " size:16" << endl;
				else
					cout << " offset:"
					     << (c.count&15)*32
					     << " size:"
					     << (32<<(c.count>>4)) << endl;
			} else 
				cout << " count:" << c.count << endl;
		}
	}
}


bool
get_next_token(string &str, string &tok)
{
	int i, p;

	for(i=0; str.chars()[i] && (str[i] <= ' ' || str[i] == '='); i++)
		continue; 
	if(!str.chars()[i] || str[i] == '#') {
		tok = "";
		return false;
	}

	p = i;
	for(; str.chars()[i] && str[i] > ' ' && str[i] != '='; i++)
		continue;
	tok = str(p, i-p);
	str = str(i, str.length()-i);
	return true;
}


int
main(int argc, csz **argv)
{
	ArgParse arg(argc, argv);
	ofstream mapf;
	string rootdir = "/";
	string destdir = "";
	string bootdev;
	string var, val, token;
	string config_fname = "boot/solo.cnf";
	string bootblock_fname, map_fname;
	string fname;
	struct MapFile mapfile;
	bool compact = false;
	bool uninstall = false;
	int bootoffset = 0;
	int mapoffset = 0;
	bool boot_override = false;

	memset(&mapfile, 0, sizeof(mapfile));
	while(arg.more()) {
		if(arg.match("-h")) {
			cerr << "usage: " << argv[0]
			<< " [-d] [-r root dir] [-b boot device] "
			"<config_file>" << endl;
			exit(1);
		} else if(arg.match("-v")) {
			cout << "mksolo v" mksolo_VER_STR " (" 
				mksolo_VER_DATE ")" << endl;
			exit(1);
		} else if(!arg.match("-r", &rootdir) && 
		   !arg.match("-b", &bootdev) &&
		   !arg.match("-u", &uninstall) &&
		   !arg.match("-d", &debug)) {
			if(arg.prefix("-"))
				arg.badoption();
			else 
				arg.get(&config_fname);
		}
	}
	if(!bootdev.length()) 
		find_device(&bootdev, rootdir.chars());
	else
		boot_override = true;

	config_fname = rootdir + "/" + config_fname; 

	ifstream f(config_fname.chars());
	if(!f.is_open()) {
		perror(config_fname.chars());
		exit(1);
	}

	while(f.good()) {
		sz buf[4096];
		int bi;	/* basename index */
		f.getline(buf, sizeof buf);
		string str = buf;
		if(!get_next_token(str, var))
			continue;
		get_next_token(str, val);
	
		bi = 0;	
		for(int i=0; i<val.length(); i++)
			if(val[i] == '/')
				bi = i+1;

		if(var == "compact")
			compact = true;
		else if(var == "bootblock")
			bootblock_fname = rootdir+"/"+val;
		else if(var == "boot") {
			if(boot_override)
				cout << "'boot' entry overridden" << endl;
			else
				bootdev = val;
		} else if(var == "saveboot") {
			if(uninstall)
				restore_boot_block(bootdev, rootdir+"/"+val);
			else
				save_boot_block(bootdev, rootdir+"/"+val);
		} else if(var == "map") {
			map_fname = rootdir+"/"+val;
			if(uninstall)
				continue;
			mapf.open(map_fname.chars());
			if(!mapf.is_open()) {
				perror(map_fname.chars());
				exit(1);
			}
		} else if(var == "dir") {
			destdir = val;
		} else if(var == "image" || var == "loader") {
			if(uninstall)
				continue;
			string dval = val;
			if(destdir.length() > 0)
				dval = destdir + "/" + dval(bi, val.length()-bi);
			while(dval[0] == '/')
				dval = dval(1, dval.length()-1);
			if(var == "loader") {
				bootoffset = mapoffset;
				if(debug)
					cout << "Loader: " << val << endl;
			}
			if(val[0] != '/')
				val = rootdir + "/" + val;
			File_Geometry geo(val, compact);
			mapfile.fsize = geo.get_size();
			mapfile.nchunks = 0;
			mapfile.device = DEV_DISK | DEVICE_MAPPED;
			mapfile.unit = geo.get_unit();
			mapfile.mtime_hi = geo.get_mtime_hi();
			mapfile.namelen = dval.length()+1;
			// streampos top = mapf.seekoff(0, ios::cur);
			streampos top = mapf.tellp();
			mapf.write(&mapfile, sizeof(mapfile));
			mapoffset += sizeof(mapfile) + mapfile.namelen;
			mapf << dval << '\0';

			Chunk c;
			while(geo.next_chunk(&c)) {
				mapf.write(&c, sizeof(c));
				mapoffset += sizeof(c);
				mapfile.nchunks++;
			}
			// streampos bot = mapf.seekoff(0, ios::cur);
			streampos bot = mapf.tellp();
			// mapf.seekpos(top);
			mapf.seekp(top);
			mapf.write(&mapfile, sizeof(mapfile));
			// mapf.seekpos(bot);
			mapf.seekp(bot);
			if(bot > MAP_SIZE)
				cerr << "mapfile too large: " << bot << " > " << MAP_SIZE << endl;
		} else if(var == "device") {
			if(uninstall)
				continue;
			if(destdir.length() > 0)
				val = destdir + "/" + val(bi, val.length()-bi);
			while(val[0] == '/')
				val = val(1, val.length()-1);
			mapfile.nchunks = 0;
			get_next_token(str, token);
			int dev = strtol(token.chars(), 0, 0);
			mapfile.fsize = 0;
			mapfile.device = dev;
			get_next_token(str, token);
			mapfile.unit  = strtol(token.chars(), 0, 0);
			mapfile.fsize = 0;
			mapfile.mtime_hi =
				((int64)time(0)*(int64)1000000000) >> 32;
			mapfile.namelen = val.length()+1;
			mapf.write(&mapfile, sizeof(mapfile));
			mapoffset += sizeof(mapfile) + mapfile.namelen;
			mapf << val << '\0';
			if(debug) 
				cout << "Device: " << hex << (int)mapfile.device
					<< "/" << mapfile.unit
					<< " file: " << val << endl;
		} else {
			cerr << "unknown configuration variable: "
				<< var << endl;
			exit(1);
		}
	}
	if(!uninstall) {
		File_Geometry geo_map(map_fname, compact);
		mapfile.namelen = 0;
		mapf.write(&mapfile, sizeof(mapfile));
		f.close();
		mapf.close();
		write_boot_block(bootdev, bootblock_fname, geo_map, bootoffset);
	}
}

